"""
pyfnalsnow - a python module to interact with the JSON API of the FNAL
Service Now (SNOW) instance.
"""

#########################################################################
### Configuration #######################################################
#########################################################################

config_file = '/etc/snow/config.yaml'
config = {}

cache = {}

modules = {
    'incident':    'pyfnalsnow.Incident',
    'sc_req_item': 'pyfnalsnow.RITM',
    'sc_request':  'pyfnalsnow.Request',
}

types = {
    'incident':   'Incident',
    'prjtask':    'Project Task',
    'sc_request': 'Request',
    'ritm':       'Requested Item',
    'task':       'Task',
}


#########################################################################
### Declarations ########################################################
#########################################################################

import pysnow, re, sys, yaml
import pyfnalsnow.Incident, pyfnalsnow.Request, pyfnalsnow.RITM

import urllib3
urllib3.disable_warnings()

#########################################################################
### Definitions #########################################################
#########################################################################

impact = {}
priority = {}
urgency = {
    '1': '1 - Critical',
    '2': '2 - High',
    '3': '3 - Medium',
    '4': '4 - Low',
}

#########################################################################
### Configuration Subroutines ###########################################
#########################################################################


def pyfnalsnow_config(file=config_file):
    """
    Load a yaml configuration from a configuration file.  Sets a global
    'config' variable.
    """
    global config

    try:
        config = yaml.load(open(file, 'r'))
    except IOError as e:
        print("file error:  %s" % e)
        sys.exit(2)
    except Exception as e:
        print("unknown error:  %s" % e)
        sys.exit(2)

    return config

#########################################################################
### pysnow wrappers #####################################################
#########################################################################

def connect():
    """
    Connect to ServiceNow with credentials from the configuration file.
    Returns the pysnow client object directly.
    """

    global snow

    snow = pysnow.Client(
        instance=config['servicenow']['instance'],
        user=config['servicenow']['username'],
        password=config['servicenow']['password']
    )

    return snow

#########################################################################
### pyfnalsnow wrappers #################################################
#########################################################################

def cacheQueryOne(table, query):
    """
    Do a SNOW database query and cache the result for later.  Uses
    'get_one()' so we only get a single result.  Especially useful for
    user and group lookups, which change rarely and need to be queried
    several times.
    """
    if table not in cache: cache[table] = {}

    q = str(query)

    if q in cache[table]:
        return cache[table][q]
    else:
        try:
            r = snow.resource(api_path='/table/%s' % table).get(
                query=query,
                stream=True
            )
            entries = []
            for record in r.all():
                entries.append(record)
            if len(entries) == 0:
                raise Exception('no matching entries (%s vs %s)'
                    % (query, table))
            elif len(entries) > 1:
                raise Exception('too many entries (%s) (%s vs %s)'
                     % (len(entries), query, table))
            result = entries[0]
        except Exception as e:  # noqa: F841
            result = {'value': '(no match)'}
        cache[table][q] = result
    return cache[table][q]

def tableSwitch(table, function, **args):
    """
    Black Magic Ahoy!  Given a function name and a table type, invokes that
    function *from the matching class*.
    """
    if table in modules:
        return eval("%s.%s" % (modules[table], function))(**args)
    else:
        raise Exception('unsupported table: %s' % type)

def typeSwitch(tkt, function, **kwargs):
    """
    Black Magic Ahoy!  Given a function name and a ticket, invokes that
    function *from the matching class*.
    """
    try:    number = tkt['number']
    except: raise Exception('no ticket number in ticket')
    type = tktType(number)

    if type in modules:
        return eval("%s.%s" % (modules[type], function))(tkt, **kwargs)
    else:
        raise Exception('unsupported type: %s' % type)

#########################################################################
### Reporting Functions #################################################
#########################################################################


def tktString(tkt, type='base', *args, **kwargs):
    """
    Create a string summarizing ticket data, and return as an array.
    """

    if type == 'audit': return tktStringBaseAudit(tkt, **kwargs)
    elif type == 'base': return tktStringBase(tkt, **kwargs)
    elif type == 'debug': return tktStringDebug(tkt, **kwargs)
    elif type == 'short': return tktStringShort(tkt, **kwargs)
    elif type == 'worklog': return tktStringWorklog(tkt, **kwargs)
    else:
        raise Exception('unknown string type: %s' % type)

def tktIsResolved(tkt): return typeSwitch(tkt, 'tktIsResolved')

def tktStringAssignee(tkt): return typeSwitch(tkt, 'tktStringAssignee')
def tktStringAudit(tkt): return typeSwitch(tkt, 'tktStringAudit')
def tktStringBase(tkt): return typeSwitch(tkt, 'tktStringBase')
def tktStringBaseAudit(tkt): return typeSwitch(tkt, 'tktStringBaseAudit')
def tktStringDebug(tkt): return typeSwitch(tkt, 'tktStringDebug')
def tktStringDescription(tkt): return typeSwitch(tkt, 'tktStringDescription')
def tktStringJournal(tkt): return typeSwitch(tkt, 'tktStringJournal')
def tktStringPrimary(tkt): return typeSwitch(tkt, 'tktStringPrimary')
def tktStringRequestor(tkt): return typeSwitch(tkt, 'tktStringRequestor')
def tktStringResolution(tkt): return typeSwitch(tkt, 'tktStringResolution')
def tktStringShort(tkt): return typeSwitch(tkt, 'tktStringShort')
def tktStringSummary(tkt): return typeSwitch(tkt, 'tktStringSummary')
def tktStringWorklog(tkt): return typeSwitch(tkt, 'tktStringWorklog')

#########################################################################
### ServiceNow Searching ################################################
#########################################################################

def ciById(id):
    """
    Pull a cmdb_ci by sys_id.  Goes through cacheQueryOne(), so
    future calls are cached.
    """
    return cacheQueryOne('cmdb_ci', query={'sys_id': id})

def ciByName(name):
    """
    Pull a cmdb_ci by name.  Goes through cacheQueryOne(), so
    future calls are cached.
    """
    return cacheQueryOne('cmdb_ci', query={'name': name})

def groupById(id):
    """
    Pull a sys_user_group entry by sys_id.  Goes through cacheQueryOne(), so
    future calls are cached.
    """
    return cacheQueryOne('sys_user_group', query={'sys_id': id})

def groupByName(name):
    """
    Pull a sys_user_group entry by name.  Goes through cacheQueryOne(), so
    future calls are cached.
    """
    return cacheQueryOne('sys_user_group', query={'name': name})

def groupLink(group):
    """
    Convert a 'group' style object to a group name.  This is the data
    structure that sometimes appears in the middle of standard results.
    """
    if isinstance(group, dict):
        if 'value' in group:
            if group['value'] == '0': return '*none*'
            g = groupById(group['value'])
            return g['name']
    return group

def tktByNumber(number):
    """
    Look up a ticket number by number.  We will look up the correct table
    type based on the ticket number as well.  Returns a single object.
    """

    type = tktType(number)

    r = snow.resource(api_path='/table/%s' % type).get(
        query={'number': number},
        stream=True
    )

    entries = []
    for record in r.all():
        entries.append(record)

    if len(entries) == 0:
        raise Exception('%s: no matching entries' % number)
    elif len(entries) > 1:
        raise Exception('%s: too many entries (%s)' % number, len(entries))
    return entries[0]

def tktHistory(tkt):
    """
    Given a full ticket, query for associated sys_history_line objects.
    """

    r = snow.resource(api_path='/table/%s' % 'sys_history_line').get(
        query={'set.id': tkt['sys_id']},
        stream=True
    )
    entries = []
    for record in r.all(): entries.append(record)

    ret = []
    for i in entries:
        ret.append(i)
    return ret

def tktPending(tkt, **kwargs):
    """
    Given a full ticket, set the status of that ticket to 'Pending'.
    """
    return typeSwitch(tkt, 'tktPending', **kwargs)

def tktResolve(tkt, update, **kwargs):
    """
    Given a full ticket, close that ticket out.
    """
    return typeSwitch(tkt, 'tktResolve', update=update, **kwargs)

def tktJournalEntries(tkt):
    """
    Given a full ticket, query for journal entries.
    """

    r = snow.resource(api_path='/table/%s' % 'sys_journal_field').get(
        query={'element_id': tkt['sys_id']}, stream=True)
    entries = []
    for record in r.all(): entries.append(record)

    ret = []
    journals = {}
    for i in entries:
        if 'sys_created_on' in i:
            key = i['sys_created_on']
            journals[key] = i

    for i in sorted(journals.keys()):
        ret.append(journals[i])
    return ret

def tktSearch(table, **args):
    """
    Given a table name, search the table using tktFilter().  Returns an
    array of matching objects.
    """

    try:
        search = tableSwitch(table, 'tktFilter', **args)
        r = snow.resource(api_path='/table/%s' % table).get(query=search,
            stream=True)
    except Exception as e:
        raise e

    ret = []
    for i in r.all():
        if i: ret.append(i)
    return ret

def tktCreate(table, entry, **args):
    """
    """
    return snow.insert(table, entry, **args)

def tktUpdate(number, updateHash):
    """
    Update a ticket number by number.  We will look up the correct table
    type based on the ticket number as well.  Returns a single object.
    """
    type = tktType(number)
    r = snow.resource(api_path='/table/%s' % type).update(
        query={'number': number},
        payload=updateHash
    )

    ret = []
    for i in r.all(): ret.append(i)
    if len(ret) >= 1: return ret[0]
    return []

def userById(id):
    """
    Pull a sys_user entry by sys_id.  Goes through cacheQueryOne(), so
    future falls are cached.
    """
    return cacheQueryOne('sys_user', query={'sys_id': id})

def userByName(name):
    """
    Pull a sys_user entry by name.  Goes through cacheQueryOne(), so
    future falls are cached.
    """
    return cacheQueryOne('sys_user', query={'name': name})

def userByUsername(name):
    """
    Pull a sys_user entry by user_name.  Goes through cacheQueryOne(), so
    future falls are cached.
    """
    return cacheQueryOne('sys_user', query={'user_name': name})

def userInGroup(username, group):
    """
    Return True is the user with the offered username is in the group with
    the offered group name, False otherwise.
    """

    groups = userInGroups(username)
    for g in groups:
        if g == group: return True
    return False

def userInGroups(username):
    """
    Given a username, return a list of group names that this user is a
    member of.
    """
    me = userByUsername(username)
    ret = []
    if me:
        r = snow.resource(api_path='/table/%s' % 'sys_user_grmember').get(
            query={'user': me['sys_id']},
            stream=True
        )
        for g in r.all():
            id = g['group']['value']
            group = groupById(id)
            ret.append(group['name'])
    return ret

def userLinkName(user):
    """
    Convert a 'user' style object to a name.  This is the data structure
    that sometimes appears in the middle of standard results.
    """
    if isinstance(user, dict):
        if 'value' in user:
            if user['value'] == '0': return '*nobody*'
            u = pyfnalsnow.userById(user['value'])
            return u['name']
    return user

def userLinkUsername(user):
    """
    Convert a 'user' style object to a user_name.  This is the data structure
    that sometimes appears in the middle of standard results.
    """
    if isinstance(user, dict):
        if 'value' in user:
            if user['value'] == '0': return '*nobody*'
            u = pyfnalsnow.userById(user['value'])
            return u['user_name']
    return user

#########################################################################
### Ticket Internals ####################################################
#########################################################################

def tktNumberParse(number):
    """
    Convert human-provided ticket numbers to the correct style for
    ServiceNow.  This means a) set the length correctly by ticket type and
    b) if we don't have a type, assume INC.

    Note that this doesn't cover the case where we get a number that is
    too long.
    """

    m = re.match('^(TASK|RITM|INC|REQ|PRJTASK)(\d+)$', number, re.IGNORECASE)
    if m:
        type = m.group(1).upper()
        num = m.group(2)
        length = len(number)
        if type == 'INC': return 'INC%s%s' % ('0' * (15 - length), num)
        if type == 'REQ': return 'REQ%s%s' % ('0' * (15 - length), num)
        if type == 'TASK': return 'TASK%s%s' % ('0' * (11 - length), num)
        if type == 'RITM': return 'RITM%s%s' % ('0' * (11 - length), num)
        if type == 'PRJTASK': return 'PRJTASK%s%s' % ('0' * (11 - length), num)
        raise 'unknown ticket type: %s' % type

    elif re.match('^(\d+)$', number):
        return "INC%s%s" % ('0' * (12 - len(number)), number)

    else:
        raise 'could not parse "%s"' % number

def tktType(number):
    """
    Given a ticket number, return the matching ServiceNow table name.
    """

    num = tktNumberParse(number)
    m = re.match('^(TASK|RITM|INC|REQ|PRJTASK)(\d+)$', num, re.IGNORECASE)
    if m:
        type = m.group(1).upper()
        if type == 'INC': return 'incident'
        if type == 'PRJTASK': return 'pm_project_task'
        if type == 'REQ': return 'sc_request'
        if type == 'RITM': return 'sc_req_item'
        if type == 'TASK': return 'sc_task'

    raise 'unknown ticket type for %s' % number

#########################################################################
### main () #############################################################
#########################################################################

pyfnalsnow_config(config_file)



"""
pyfnalsnow.RITM
"""

#########################################################################
### Declarations ########################################################
#########################################################################

import pyfnalsnow   # noqa:F401
from pyfnalsnow.ticket import tktStringAssignee     # noqa:F401
from pyfnalsnow.ticket import tktStringAudit        # noqa:F401
from pyfnalsnow.ticket import tktStringBase         # noqa:F401
from pyfnalsnow.ticket import tktStringBaseAudit    # noqa:F401
from pyfnalsnow.ticket import tktStringDebug        # noqa:F401
from pyfnalsnow.ticket import tktStringDescription  # noqa:F401
from pyfnalsnow.ticket import tktStringJournal      # noqa:F401
from pyfnalsnow.ticket import tktStringPrimary      # noqa:F401
from pyfnalsnow.ticket import tktStringRequestor    # noqa:F401
from pyfnalsnow.ticket import tktStringResolution   # noqa:F401
from pyfnalsnow.ticket import tktStringShort        # noqa:F401
from pyfnalsnow.ticket import tktStringSummary      # noqa:F401

#########################################################################
### Configuration #######################################################
#########################################################################

state = {
    -5: 'Pending',
     3: 'Closed',       # noqa:E121
    11: 'Complete',
    20: 'Waiting for Acknowledgement',
    21: 'Acknowledged',
    23: 'Work in Progress',
    24: 'Cancelled'
}


#########################################################################
### Subroutines #########################################################
#########################################################################

def tktFilter(status='open', **args):
    """
    Filter tickets.
    """

    extra = []
    if status.lower() == 'open':            # not pending, closed, cancelled, complete
        extra.append('state!=-5')
        extra.append('state!=3')
        extra.append('state!=24')
        extra.append('state!=11')
    elif status.lower() == 'closed':
        extra.append('state==3')
    elif status.lower() == 'unresolved':    # not closed, cancelled, complete
        extra.append('state!=3')
        extra.append('state!=24')
        extra.append('state!=11')

    if 'unassigned' in args:
        extra.append('assigned_to=NULL')

    if 'group' in args:
        group = pyfnalsnow.groupByName(args['group'])
        extra.append('assignment_group=%s' % group['sys_id'])

    if 'assigned' in args:
        user = pyfnalsnow.userByUsername(args['assigned'])
        extra.append('assigned_to=%s' % user['sys_id'])

    if 'submit' in args:
        user = pyfnalsnow.userByUsername(args['submit'])
        extra.append('sys_created_by=%s' % user['user_name'])

    search = '^'.join(extra)

    return search


def tktIsResolved(tkt):
    """
    Returns True if the RITM is resolved, False otherwise.
    """
    if tkt['state'] == '3': return True         # closed
    if tkt['state'] == '24': return True        # cancelled
    if tkt['state'] == '23': return True        # work in progress?  huh?
    if tkt['state'] == '11': return True        # Complete
    else: return False

def tktPending(tkt, **kwargs):
    """
    Set the state of this RITM to 'pending' (-5).  As with tktResolve(),
    this is complicated because you have to move from one state to
    another in a very specific manner.

    kwargs:
      debug     Boolean; default False
      reason    String; default is 'Customer'
    """

    goal = -5

    try:    debug = kwargs['debug']
    except: debug = False

    try:    reason = kwargs['reason']
    except: reason = 'Customer'

    new = {}

    state = tkt['state']
    if debug: print("initial state: %s  goal state: %s" % (state, goal))

    if int(state) == int(goal):
        if debug: print("nothing to do (already in goal state)")
        return tkt

    if int(state) == -5: raise Exception('at goal state')
    elif int(state) == 3: raise Exception('cannot change state from 3')
    elif int(state) == 11: raise Exception('cannot change state from 11')
    elif int(state) == 20: new['state'] = 21
    elif int(state) == 21: new['state'] = 23
    elif int(state) == 23: new['state'] = -5
    elif int(state) == 24: raise Exception('cannot change state from 24')
    else: raise Exception('invalid start state: %s' % state)

    if debug: print("updating goal state to %s" % new['state'])

    if new['state'] == int(goal):
        if debug: print("going to goal state, will also set u_pending_reason")
        new['u_pending_reason'] = reason

    updated = pyfnalsnow.tktUpdate(tkt['number'], new)
    return tktPending(updated, **kwargs)


def tktResolve(tkt, update, **kwargs):
    """
    Resolve an RITM.  This is complicated because there are a very
    specific list of state changes for the 'state' field that are
    acceptable.  The goal must be either 3 (closed, default) or 24
    (cancelled).

    update is a dict of key/value pairs that will be used for making
    updates.  You must pass in 'text' (will be used as close_notes).

    kwargs:
      debug     Boolean, default False
      goal      Integer, default 3, must be either 3 or 24.

    """

    try:    goal = int(kwargs['goal'])
    except: goal = 3

    if goal == 3: pass
    elif goal == 24: pass
    else: raise Exception('invalid goal state: %d' % goal)

    try: text = update['text']
    except: raise Exception('required field: text')

    try:    debug = kwargs['debug']
    except: debug = False

    new = {}

    ###########################
    ### valid state changes ###
    ###########################

    # -5 -> 23, 24
    #  3 -> no way out
    # 11 -> no way out
    # 20 -> 21, 24(?)
    # 21 -> 23, 24
    # 23 -> -5, 3, 24
    # 24 -> no way out

    state = tkt['state']

    if debug: print("initial state: %s  goal state: %s" % (state, goal))

    if int(state) == int(goal):
        if debug: print("nothing to do (already in goal state)")
        return tkt

    if int(state) == -5:
        if goal == 3: new['state'] = 23
        elif goal == 24: new['state'] = 24
        else: new['state'] = goal
    elif int(state) == 3: raise Exception('cannot change state from 3')
    elif int(state) == 11: raise Exception('cannot change state from 11')
    elif int(state) == 20:
        if goal == 3: new['state'] = 21
        elif goal == 24: new['state'] = 24
        else: new['state'] = goal
    elif int(state) == 21:
        if goal == 3: new['state'] = 23
        elif goal == 24: new['state'] = 24
        else: raise Exception('not a valid goal for resolution')
    elif int(state) == 23:
        if goal == 3: new['state'] = 3
        elif goal == 24: new['state'] = 24
        else: raise Exception('not a valid goal for resolution')
    elif int(state) == 24: raise Exception('cannot change state from 24')
    else:
        raise Exception('invalid start state: %s' % state)

    if debug:
        print("updating goal state to %s" % new['state'])

    if new['state'] == int(goal):
        if debug: print("going to goal state, will also set close_notes")
        new['close_notes'] = text

    updated = pyfnalsnow.tktUpdate(tkt['number'], new)
    return tktResolve(updated, update, **kwargs)



from setuptools import setup
import glob, re, os

## get documentation from README.md
with open("README.md", "r") as fh:
    long_description = fh.read()

## get version from spec file
with open('fnal-snow-python.spec', 'r') as fh:
    for line in fh:
        m = re.search("^Version:\s+(.*)\s*$", line)
        if m:
            version=m.group(1)
            break

## get list of files to install
pyfiles = glob.glob(os.path.join('*', '*.py'))
pyfiles = [pyfile[:-3] for pyfile in pyfiles]

scripts = glob.glob(os.path.join('usr/sbin/*'))
man     = glob.glob(os.path.join('man/man1/*'))

setup (
  author_email = 'tskirvin@fnal.gov',
  author = 'Tim Skirvin',
  data_files = [ ( 'share/man/man1', man ) ],
  description = 'SNOW JSON API access',
  license = 'Perl Artistic',
  install_requires = ['pysnow>=0.7.4', 'PyYAML>=3.11'],
  keywords = ['snow', 'service-now'],
  long_description_content_type = 'text/markdown',
  long_description = long_description,
  maintainer_email = 'tskirvin@fnal.gov',
  maintainer = 'Tim Skirvin',
  name = 'pyfnalsnow',
  package_dir = { 'pyfnalsnow': 'pyfnalsnow' },
  py_modules = pyfiles,
  scripts = scripts,
  url = 'https://github.com/tskirvin/pyfnalsnow.git',
  version = version
)
